import logging
from typing import Callable, Any, Optional, NoReturn
from dataclasses import dataclass
from queue import Queue

from telebox.utils.thread_pool import ThreadPool


logger = logging.getLogger(__name__)


@dataclass
class Task:
    task: Callable
    args: tuple
    kwargs: dict[str, Any]


class TaskExecutor:

    def __init__(self, threads: int):
        self._thread_pool = ThreadPool(threads, self._run_task_processing)
        self._tasks = Queue()

    def add_task(
        self,
        task: Callable,
        args: tuple = (),
        kwargs: Optional[dict[str, Any]] = None
    ) -> None:
        task = Task(
            task=task,
            args=args,
            kwargs=kwargs or {}
        )
        logger.debug("Task added to queue: %r.", task)
        self._tasks.put(task, block=False)

    def run_tasks(self) -> None:
        logger.debug("Tasks is starting...")
        self._thread_pool.start()
        logger.info("Tasks started.")

    def wait_tasks(self) -> None:
        logger.info("Tasks finishing...")
        self._tasks.join()
        logger.info("Tasks finished.")

    def _run_task_processing(self) -> NoReturn:
        while True:
            task = self._tasks.get()

            # noinspection PyBroadException
            try:
                logger.debug("Task processing started: %r.", task)
                task.task(*task.args, **task.kwargs)
            except Exception:
                logger.exception("An error occurred while processing a task!")
            finally:
                self._tasks.task_done()
                logger.debug("Task processing finished: %r.", task)



from dataclasses import dataclass

from telebox.utils.env.errors import EnvError


@dataclass
class Env:

    def check(self) -> None:
        for name, value in vars(self).items():
            if value is None:
                raise EnvError(f"{name!r} is not set!")



from typing import Type

from telebox.dispatcher.filters.error_filter import AbstractErrorFilter
from telebox.dispatcher.typing import Event


class ErrorTypeFilter(AbstractErrorFilter):

    def __init__(self, *types: Type[Exception], strictly: bool = False):
        if not types:
            raise ValueError("No types!")

        self._types = types
        self._strictly = strictly

    def get_value(self, error: Exception, event: Event) -> Exception:
        return error

    def check_value(self, value: Exception) -> bool:
        return type(value) in self._types if self._strictly else isinstance(value, self._types)



from telebox.dispatcher.middlewares.middleware import Middleware
from telebox.dispatcher.typing import Event
from telebox.dispatcher.enums.event_type import EventType
from telebox.utils.callback_data_builders.builder import AbstractCallbackDataBuilder


class CallbackDataMiddleware(Middleware):

    def __init__(self, builder: AbstractCallbackDataBuilder):
        self._builder = builder

    def process_event(self, event: Event, event_type: EventType) -> None:
        if event_type is EventType.CALLBACK_QUERY:
            _, event.data = self._builder.parse(event.data)



from .dispatcher import Dispatcher, Event
from .enums import EventType, MediaGroupContentType
from .handlers import AbstractEventHandler, AbstractErrorHandler
from .filters import AbstractEventFilter, AbstractErrorFilter
from .middlewares import Middleware
from .utils import MediaGroup, RateLimit, Limit, get_event_chat_id, get_event_user_id


__all__ = [
    "Dispatcher",
    "Event",
    "EventType",
    "MediaGroupContentType",
    "AbstractEventHandler",
    "AbstractErrorHandler",
    "AbstractEventFilter",
    "AbstractErrorFilter",
    "Middleware",
    "MediaGroup",
    "RateLimit",
    "Limit",
    "get_event_chat_id",
    "get_event_user_id"
]



from .not_set import NotSet
from .named_set import NamedSet
from .callback_data_builders import AbstractCallbackDataBuilder
from .task_executor import TaskExecutor
from .env import Env


__all__ = [
    "NotSet",
    "NamedSet",
    "AbstractCallbackDataBuilder",
    "TaskExecutor",
    "Env"
]



from typing import Optional, Union

from telebox.dispatcher.filters.event_filter import AbstractEventFilter
from telebox.dispatcher.enums.event_type import EventType
from telebox.dispatcher.utils.media_group import MediaGroup
from telebox.telegram_bot.types.types.message import Message
from telebox.telegram_bot.types.types.inline_query import InlineQuery
from telebox.telegram_bot.types.types.chosen_inline_result import ChosenInlineResult
from telebox.telegram_bot.types.types.callback_query import CallbackQuery
from telebox.telegram_bot.types.types.shipping_query import ShippingQuery
from telebox.telegram_bot.types.types.pre_checkout_query import PreCheckoutQuery
from telebox.telegram_bot.types.types.poll_answer import PollAnswer
from telebox.telegram_bot.types.types.chat_member_updated import ChatMemberUpdated
from telebox.telegram_bot.types.types.chat_join_request import ChatJoinRequest


class UserIDFilter(AbstractEventFilter):

    def __init__(self, *ids: int):
        self._ids = set(ids)

    def get_event_types(self) -> set[EventType]:
        return {
            EventType.MESSAGE,
            EventType.EDITED_MESSAGE,
            EventType.CHANNEL_POST,
            EventType.EDITED_CHANNEL_POST,
            EventType.MEDIA_GROUP,
            EventType.INLINE_QUERY,
            EventType.CHOSEN_INLINE_RESULT,
            EventType.CALLBACK_QUERY,
            EventType.SHIPPING_QUERY,
            EventType.PRE_CHECKOUT_QUERY,
            EventType.POLL_ANSWER,
            EventType.MY_CHAT_MEMBER,
            EventType.CHAT_MEMBER,
            EventType.CHAT_JOIN_REQUEST
        }

    def get_value(
        self,
        event: Union[Message,
                     MediaGroup,
                     InlineQuery,
                     ChosenInlineResult,
                     CallbackQuery,
                     ShippingQuery,
                     PreCheckoutQuery,
                     PollAnswer,
                     ChatMemberUpdated,
                     ChatJoinRequest]
    ) -> Optional[int]:
        return event.user_id

    def check_value(self, value: Optional[int]) -> bool:
        return value in self._ids if self._ids else value is not None



from dataclasses import dataclass
from typing import Optional

from telebox.telegram_bot.types.type import Type


@dataclass(unsafe_hash=True)
class PhotoSize(Type):
    file_id: str
    file_unique_id: str
    width: int
    height: int
    file_size: Optional[int] = None
